/* Copyright (C) 2024 GSI Helmholtzzentrum fuer Schwerionenforschung, Darmstadt
   SPDX-License-Identifier: GPL-3.0-only
   Authors: Volker Friese [committer] */

#pragma once

#include "CbmDefs.h"

#include <map>

#include <yaml-cpp/yaml.h>


namespace cbm::algo::evbuild
{

  /** @class V0TriggerConfig
   ** @brief Configuration of the V0 trigger class (trigger on displaced vertices)
   ** @author Volker Friese <v.friese@gsi.de>
   ** @since 19.04.2024
   **/
  class V0TriggerConfig {

   public:
    /** @brief Default constructor **/
    V0TriggerConfig(){};

    /** @brief Constructor with parameters
     ** @param cutTime  Maximum time difference of tracks [ns]
     ** @param cutZ     Minimum z position at closest approach
     ** @param cutDist  Maximum distance at closest approach
     **/
    V0TriggerConfig(double cutTime, double cutZ, double cutDist)
      : fCutTime(cutTime)
      , fCutZ(cutZ)
      , fCutDist(cutDist)
      , fIsSet(true)
    {
    }

    /** @brief Constructor from YAML **/
    V0TriggerConfig(YAML::Node config);

    /** @brief Destructor **/
    ~V0TriggerConfig() = default;

    /** @brief Cut on time difference between tracks **/
    double CutTime() const { return fCutTime; }

    /** @brief Cut on z position of closest approach **/
    double CutZ() const { return fCutZ; }

    /** @brief Cut on distance of tracks at closest approach **/
    double CutDist() const { return fCutDist; }

    /** @brief Check whether config was set **/
    bool IsSet() const { return fIsSet; }

    /** @brief Save to YAML **/
    YAML::Node ToYaml() const;


   private:
    double fCutTime{0.};  /// Maximum time difference of tracks
    double fCutZ{0.};     /// Minimum z position at closest approach
    double fCutDist{0.};  /// Maximum distance at closest approach
    bool fIsSet{false};   /// Flag whether a configuration was set
  };


}  // namespace cbm::algo::evbuild
